﻿// Copyright (c) 2014, Leandro Toss Hoffmann (INPE)> <leandro.hoffmann@inpe.br >
// Permission to use, copy, modify, and distribute this model for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
// This model is provided "as is" and the author disclaims all warranties with regard to it. 

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using EA;

namespace SMP2CodeGen
{
    class AssemblyGenClass
    {
        string path = "";
        EA.Repository Repository;
        Diagram diagram;
        System.IO.StreamWriter file;

        public void GenerateAssembly(EA.Repository a_Repository, string Location)
        {
            Repository = a_Repository;
            diagram = Repository.GetCurrentDiagram();

            if (diagram == null)
            {
                MessageBox.Show("Please, select a Diagram from the Tree View");
                return;
            }

            // Gets the path to write files
            CodeGenDialog dialog = new CodeGenDialog();
            DialogResult result = dialog.ShowDialog();
            if (result == DialogResult.OK)
                path = dialog.GetPath();
            else
                return;

            file = new System.IO.StreamWriter(@path + "\\" + diagram.Name.Trim() + ".asb", false);

            String idents = "";

            // Assembly Header
            string datetime = DateTime.Now.ToString("yyyy-MM-ddTHH:mm:ss+HH:mm");
            file.WriteLine("<?xml version=\"1.0\" encoding=\"utf-8\"?>");
            file.WriteLine("<Assembly:Assembly Id=\"" + diagram.Name.Trim() + "\" Name=\"RootModel\" Date=\"" + datetime +  "\" Creator=\"Simubox's SMP2 Assembly Generator\" Version=\"1.0\"");
            file.WriteLine("Implementation=\"10000000-0000-0001-0001-e2d5fd9a0d2c\"");
            file.WriteLine("xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\"");
            file.WriteLine("xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
            file.WriteLine("xmlns:xlink=\"http://www.w3.org/1999/xlink\"");
            file.WriteLine("xmlns:Types=\"http://www.esa.int/2008/02/Core/Types\"");
            file.WriteLine("xmlns:Elements=\"http://www.esa.int/2008/02/Core/Elements\"");
            file.WriteLine("xmlns:Catalogue=\"http://www.esa.int/2008/02/Smdl/Catalogue\"");
            file.WriteLine("xmlns:Assembly=\"http://www.esa.int/2008/02/Smdl/Assembly\">");
            idents += ("\t");
            file.WriteLine(idents + "<Description>Root Model to hold unlimited number of IModels</Description>");
            file.WriteLine(idents + "<Model xlink:title=\"RootModel\" xlink:href=\"../../models/util/SbUtil/SbUtil.cat#SbUtil.RootModel\" />");
            
            
            foreach (EA.DiagramObject obj in diagram.DiagramObjects)
            {
                Element element = Repository.GetElementByID(obj.ElementID);
                
                // Checks if the model is eligible to be a root instance
                // Is a object instance, has a name and is not contained by other objet
                if (element.Type != "Object" || element.Name == "" || IsRootModelElement(element) == false)
                    continue;          

                // Class element of instantiated object
                Element elementClass = Repository.GetElementByID(element.ClassifierID);

                file.WriteLine(idents+"<ModelInstance Id=\"" + elementClass.Name + "." + element.Name + "\" Name=\"" + element.Name + "\" Implementation=\"" + elementClass.ElementGUID.TrimStart('{').TrimEnd('}') + "\">");
                idents +=("\t");
                ProcessModelInstance(element, elementClass, "ModelsContainer", idents); // RootModel's container. This should be more flexible in the future #??#
                idents = idents.Remove(0, 1);
                file.WriteLine(idents+"</ModelInstance>");
            }
            idents = idents.Remove(0, 1);
            file.WriteLine("</Assembly:Assembly>");

            file.Close();
        }
        private void ProcessModelInstance(Element element, Element elementClass, string container, String idents)
        {
            string packageName = Repository.GetPackageByID(elementClass.PackageID).Name;

            if(element.Notes != "")
                file.WriteLine(idents+"<Description>" + element.Notes + "</Description>");
            file.WriteLine(idents+"<Model xlink:title=\"" + elementClass.Name + "\" xlink:href=\"./" + packageName + ".cat#" + elementClass.Name + "\" />");
            ProcessComposite(element, idents);
            ProcessFieldValues(element, elementClass, packageName, idents);
            ProcessFieldLinks(element, elementClass, packageName, idents);
            // #??# Process Event Links (TODO)
            ProcessReferences(element, elementClass, packageName, idents);
            file.WriteLine(idents+"<Container xlink:title=\"" + container + "\" xlink:href=\"#??#\"/>");
        }
        private void ProcessComposite(Element element, string idents)
        {
            foreach (Connector conn in element.Connectors)
            {
                if (conn.Stereotype != "SMP2composite" || conn.Type != "Aggregation")
                    continue;
                Element client = Repository.GetElementByID(conn.ClientID);
                if (client == null || client.Name == element.Name)
                    continue;

                // Checks if the models is contained by only one object
                if (HasElementSingleParent(client) == false)
                {
                    string errorMsg = "!!!!!!!!!!!!!!!!!ERROR: element " + client.Name + " has multiple parents!";
                    file.WriteLine(errorMsg);
                    MessageBox.Show(errorMsg);
                    return;
                }

                // container name
                string container = conn.ClientEnd.Role;
                // Class element of client object
                Element clientClass = Repository.GetElementByID(client.ClassifierID);
                file.WriteLine(idents + "<ModelInstance Id=\"" + clientClass.Name + "." + client.Name + "\" Name=\"" + client.Name + "\" Implementation=\"" + clientClass.ElementGUID.TrimStart('{').TrimEnd('}') + "\">");
                idents += ("\t");
                ProcessModelInstance(client, clientClass, container, idents);
                idents.Remove(0, 1);
                file.WriteLine(idents + "</ModelInstance>");
            }
        }
        private void ProcessFieldValues(Element element, Element elementClass, string packageName, String idents)
        {
            foreach (EA.Attribute att in elementClass.Attributes)
            {
                if (att.Default == "")
                    continue;

                // If there is a default value for the instance, skip the value from the model
                string defaultValue = UtilClass.GetTagValue(element.TaggedValues, att.Name);
                if (defaultValue != "")
                    continue;
                
                defaultValue = att.Default;

                if (att.IsCollection && att.UpperBound != "")
                {
                    file.WriteLine(idents + "<FieldValue xsi:type=\"" + att.Type.Replace("Smp::", "Types:") + "ArrayValue\" Field=\"./" + packageName + ".cat#" + elementClass.Name + "." + att.Name + "\">");
                    string[] values = defaultValue.Split(',');
                    int i = 0;
                    idents += ("\t");
                    foreach (string value in values)
                    {
                        file.WriteLine(idents + "<ItemValue Field=\"" + att.Name + "[" + i + "]" + "\" Value=\"" +  value.TrimStart('{').TrimEnd('}').Replace("\"", "") + "\"/>");
                        i++;
                    }
                    idents = idents.Remove(0,1);
                    file.WriteLine(idents + "</FieldValue>");
                } else
                    file.WriteLine(idents + "<FieldValue xsi:type=\"" + att.Type.Replace("Smp::", "Types:") + "Value\" Field=\"./" + packageName + ".cat#" + elementClass.Name + "." + att.Name + "\" Value=\"" + defaultValue.Replace("\"", "") + "\"/>");
            }
            // Takes the default value of the instance, if specified by a Object's tagged value
            foreach (TaggedValue tag in element.TaggedValues)
            {
                // Checks if the tag exist as a model's attribute
                bool found = false;
                bool array = false;
                string attType="";
                foreach (EA.Attribute att in elementClass.Attributes)
                {
                    if (att.Name == tag.Name)
                    {
                        found = true;
                        attType = att.Type;
                        if (att.IsCollection && att.UpperBound != "")
                            array = true;
                        break;
                    }
                }
                if (!found)
                    continue;

                string defaultValue = tag.Value;

                if (array)
                {
                    file.WriteLine(idents + "<FieldValue xsi:type=\"" + attType.Replace("Smp::", "Types:") + "ArrayValue\" Field=\"./" + packageName + ".cat#" + elementClass.Name + "." + tag.Name + "\">");
                    string[] values = defaultValue.Split(',');
                    int i = 0;
                    idents += ("\t");
                    foreach (string value in values)
                    {
                        file.WriteLine(idents + "<ItemValue Field=\"" + tag.Name + "[" + i + "]" + "\" Value=\"" + value.TrimStart('{').TrimEnd('}').Replace("\"", "") + "\"/>");
                        i++;
                    }
                    idents = idents.Remove(0, 1);
                    file.WriteLine(idents + "</FieldValue>");
                }
                else
                    file.WriteLine(idents + "<FieldValue xsi:type=\"" + attType.Replace("Smp::", "Types:") + "Value\" Field=\"./" + packageName + ".cat#" + elementClass.Name + "." + tag.Name + "\" Value=\"" + defaultValue.Replace("\"", "") + "\"/>");
            }
        }
        private void ProcessFieldLinks(Element element, Element elementClass, string packageName, String idents)
        {
            foreach (Element port in element.EmbeddedElements)
            {
                // Only considerer ports without exposed interfaces
                if (port.Type != "Port" || port.EmbeddedElements.Count > 0 || port.Stereotype == "SMP2interface")
                    continue;
                
                //int fieldId = 0;
                foreach (Connector conn in port.Connectors)
                {
                    if (conn.Stereotype != "SMP2fieldLink")
                        continue;
                    // Gets the reference to the source's port
                    Element outputField = Repository.GetElementByID(conn.ClientID); // EA names the source as client, but in this case, the client is the output field
                    if (outputField == null || outputField.ParentID == 0 || outputField.Type != "Port" || outputField.Name == port.Name)
                        continue;
                    
                    // And finally gets the reference to the port's object (Object)
                    Element source = Repository.GetElementByID(outputField.ParentID);
                    if (source == null || source.Type != "Object")
                        continue;

                    // Check if the link is based on input and onput fields
                    // #??# TODO: (1) get port's object; (2) then object's classifier; (3) then classifier's port; (4) then port's tag?
                    if (false)
                    {
                        string errorMsg = "!!!!!!!!!!!!!!!!!ERROR: fails to link fields " + outputField.Name + " -> " + port.Name + "!";
                        file.WriteLine(errorMsg);
                        MessageBox.Show(errorMsg);
                        return;
                    }
                    
                    string instanceName = element.Name;
                    string sourceName = source.Name;
                    Element sourceClass = Repository.GetElementByID(source.ClassifierID);
                    string sourceClassName = sourceClass.Name;
                    string sourcePackage = Repository.GetPackageByID(sourceClass.PackageID).Name;
                    string inputFieldName = port.Name;
                    string outputFieldName = outputField.Name;
                    //fieldId++;

                    //file.WriteLine(idents + "<Link xsi:type=\"Assembly:FieldLink\" Id=\"" + instanceName + "." + inputFieldName + fieldId + "\" Name=\"" + inputFieldName + fieldId + "\">");
                    file.WriteLine(idents + "<Link xsi:type=\"Assembly:FieldLink\" Id=\"" + instanceName + "." + inputFieldName + "\" Name=\"" + instanceName + "_" + inputFieldName + "\">");
                    idents += "\t";
                    {
                        string descr = conn.Notes;
                        if(descr != "")
                            file.WriteLine(idents + "<Description>" + descr + "</Description>");

                        file.WriteLine(idents + "<Input xlink:title=\"" + inputFieldName + "\" xlink:href=\"./" + packageName + ".cat#" + packageName + "." + elementClass.Name + "." + inputFieldName + "\" />");
                        file.WriteLine(idents + "<Source xlink:title=\"" + sourceName + "\" xlink:href=\"#" + sourceName + "\" />");
                        file.WriteLine(idents + "<Output xlink:title=\"" + outputFieldName + "\" xlink:href=\"./" + sourcePackage + ".cat#" + sourcePackage + "." + sourceClassName + "." + outputFieldName + "\" />");
                    }
                    idents = idents.Remove(0, 1);
                    file.WriteLine(idents + "</Link>");
                }
            }
        }
        private void ProcessReferences(Element element, Element elementClass, string packageName, String idents)
        {
            foreach (Element port in element.EmbeddedElements)
            {
                // Only considerer ports with Required SMP2interfaces
                if (port.Type != "Port" || port.EmbeddedElements.Count <= 0 || port.Stereotype != "SMP2interface")
                    continue;
                foreach (Element reqInt in port.EmbeddedElements)
                {
                    if (reqInt.Type != "RequiredInterface")
                        continue;

                    string container = port.Name;
                    string objInstance = element.Name;
                    int containerId = 0;

                    foreach (Connector conn in reqInt.Connectors)
                    {
                        if (conn.Stereotype != "SMP2interfaceLink")
                            continue;

                        // Gets the reference to the assembly's target (Provided Interface)
                        Element e = Repository.GetElementByID(conn.SupplierID);
                        if (e == null || e.ParentID == 0 || e.Type != "ProvidedInterface")
                            continue;

                        // Gets the reference to the interface's port (Port)
                        e = Repository.GetElementByID(e.ParentID);
                        if (e == null || e.ParentID == 0 || e.Type != "Port")
                            continue;

                        // And finally gets the reference to the port's object (Object)
                        e = Repository.GetElementByID(e.ParentID);
                        if (e == null || e.Type != "Object")
                            continue;

                        string provider = e.Name;
                        containerId++;
                        file.WriteLine(idents + "<Link xsi:type=\"Assembly:InterfaceLink\" Id=\"" + objInstance + "." + container + containerId + "\" Name=\"" + container + containerId + "\">");
                        idents += "\t";
                        {
                            string descr = conn.Notes;
                            if (descr != "")
                                file.WriteLine(idents + "<Description>" + descr + "</Description>");

                            file.WriteLine(idents + "<Reference xlink:title=\"" + container + "\" xlink:href=\"./" + packageName + ".cat#" + packageName + "." + elementClass.Name + "." + container + "\" />");
                            file.WriteLine(idents + "<Provider xlink:title=\"" + provider + "\" xlink:href=\"#" + provider + "\" />");
                        }
                        idents = idents.Remove(0, 1);
                        file.WriteLine(idents + "</Link>");
                    }
                }
            }
        }
        private bool IsRootModelElement(Element element)
        {
            // Assumes that a root model can be a client in a composite relation (i.e. a root model is not contained by other model).
            foreach (Connector conn in element.Connectors)
            {
                if (conn.Stereotype != "SMP2composite" || conn.Type != "Aggregation")
                    continue;

                Element client = Repository.GetElementByID(conn.ClientID);
                if (client == null)
                    continue;
                if (client.Name == element.Name)
                    return false;
            }
            return true;
        }
        private bool HasElementSingleParent(Element element)
        {
            int parents = 0;
            foreach (Connector conn in element.Connectors)
            {
                if (conn.Stereotype != "SMP2composite" || conn.Type != "Aggregation")
                    continue;

                Element client = Repository.GetElementByID(conn.ClientID);
                if (client == null)
                    continue;
                if (client.Name == element.Name)
                    parents ++;
            }
            if(parents > 1)
                return false;

            return true;
        }

    } // end class
} //  end namespace
