﻿///////////////////////////////////////////////////////////////////////////////
//  INSTITUTO NACIONAL DE PESQUISAS ESPACIAIS
//  Simubox - SMP2 Model Development Kit
///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2014, Leandro Toss Hoffmann (INPE)> <leandro.hoffmann@inpe.br >
// Permission to use, copy, modify, and distribute this model for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
// This model is provided "as is" and the author disclaims all warranties with regard to it. 
///////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using EA;

namespace SMP2CodeGen
{
    public class CppGenClass
    {
        private
            System.IO.StreamWriter file;
            string path = "";
            EA.Repository Repository;
            Dictionary<string, string> ParameterDirectionKind = new Dictionary<string, string> {
                        {"in", "Smp::Publication::PDK_In"}, {"out", "Smp::Publication::PDK_Out"}, {"inout", "Smp::Publication::PDK_InOut"}, {"return", "Smp::Publication::PDK_Return"}  };
            Dictionary<string, string> ViewKind = new Dictionary<string, string> {
                        {"none", "Smp::VK_None"}, {"debug", "Smp::VK_Debug"}, {"expert", "Smp::VK_Expert"}, {"all", "Smp::VK_All"} };

        public void GenerateCode(EA.Repository a_Repository, string Location)
        {
            Repository = a_Repository;

            if(Repository.GetTreeSelectedItemType() != EA.ObjectType.otElement
                && Repository.GetTreeSelectedItemType() != EA.ObjectType.otPackage)
            {
                MessageBox.Show("Please, select a Package or Class Element from the Tree View");
                return;
            }

            // Gets the path to write files
            CodeGenDialog dialog = new CodeGenDialog();
            DialogResult result = dialog.ShowDialog();
            if (result == DialogResult.OK)
            {
                path = dialog.GetPath();
            } else 
            {
                return;
            }

            // Check the type of object to be processed
            if(Repository.GetTreeSelectedItemType() == EA.ObjectType.otPackage)
            {
                ProcessPackage(Repository.GetTreeSelectedPackage());
            } else if(Repository.GetTreeSelectedItemType() == EA.ObjectType.otElement)
            {
                object eaObj;
                ObjectType objType = Repository.GetTreeSelectedItem(out eaObj);
                switch(objType)
                {
                    case ObjectType.otElement:
                        Element e = eaObj as Element;
                        if(e.Type == "Class")
                        {
                            file = new System.IO.StreamWriter(@path + "\\" + e.Name + ".cpp", false, Encoding.GetEncoding(1252));
                            //System.IO.File.Create(fileName);
                            try {
                                ProcessClass(e);
                            } catch {
                                MessageBox.Show("Error when attempting to generate source code!");
                                file.Close();
                                return;
                            }
                            file.Close();
                        } else
                            MessageBox.Show("Please, select a Class type in order to generate code.");
                        break;
                    default:
                        MessageBox.Show("Please, select a Class type in order to generate code.");
                        return;
                }
            }
        }
        private void ProcessPackage(Package package)
        {
            foreach(Element e in package.Elements)
            {
                // Check the type of object to be processed
                if(e.Type == "Package")
                    ProcessPackage(e as Package);
                else if(e.Type == "Class")
                {
                    file = new System.IO.StreamWriter(@path + "\\" + e.Name + ".cpp", false, Encoding.GetEncoding(1252));
                    try {
                        ProcessClass(e);
                    } catch {
                        MessageBox.Show("Error when attempting to generate source code!");
                        file.Close();
                        return;
                    }
                    file.Close();
                }
            }
        }
        private void ProcessClass(Element element)
        {
            string packageName = Repository.GetPackageByID(element.PackageID).Name;

            file.WriteLine("#include \"" + element.Name + ".h\"");
            file.WriteLine("#include \"Sb/EntryPoint.h\"");
            file.WriteLine("");

            string uuid = element.ElementGUID.TrimStart('{').TrimEnd('}');
            file.WriteLine("// Unique identifier of type " + element.Name);
            file.WriteLine("// UUID: " + uuid);
            file.WriteLine("const SB::Uuid " 
                + packageName 
                + "::Uuid_" 
                + element.Name 
                + "(\"" + uuid + "\");");
            file.WriteLine("");
            
            ProcessConstructorsDestructor(element, packageName);
            ProcessInitFinalise(element, packageName);
            ProcessOperations(element, packageName, "operations");
            file.WriteLine("");
            file.WriteLine("");
            ProcessOperations(element, packageName, "SMP2entryPoint");
            ProcessIModelMethods(element, packageName);
            ProcessImplementedInterfaces(element, packageName);
        }
        private void ProcessConstructorsDestructor(Element element, string packageName)
        {
            // Default Constructor
            file.WriteLine(packageName + "::" + element.Name + "::" + element.Name + "()");
            file.WriteLine("{");
            file.WriteLine("\t_Init();");
            file.WriteLine("}");
            file.WriteLine("");

            // Managed Model Constructor
            file.WriteLine(packageName + "::" + element.Name + "::" + element.Name + "(Smp::String8 name, Smp::String8 description, Smp::IComposite* parent)");
            file.WriteLine("\tthrow (Smp::Management::IManagedObject::InvalidObjectName)");
            file.WriteLine("  : SB::Management::ManagedModel(name, description, parent)");
            file.WriteLine("{");
            file.WriteLine("\t_Init();");
            file.WriteLine("}");
            file.WriteLine("");

            // Destructor
            file.WriteLine(packageName + "::" + element.Name + "::~" + element.Name + "()");
            file.WriteLine("{");
            file.WriteLine("\ttry {");
            file.WriteLine("\t\t_Destroy();");
            file.WriteLine("\t} catch (...) {");
            file.WriteLine("\t\tm_simulator->GetLogger()->Log(");
            file.WriteLine("\t\tthis,");
            file.WriteLine("\t\t\"Exception during destruction of model 'SASModel'\",");
            file.WriteLine("\t\t::Smp::Services::LMK_Error);");
            file.WriteLine("\t}");
            file.WriteLine("}");
        }
        private void ProcessInitFinalise(Element element, string packageName)
        {
            /////////// _Init()
            file.WriteLine("void " + packageName + "::" + element.Name + "::_Init()");
            file.WriteLine("{");
            file.WriteLine("");

            // Entry Points
            file.WriteLine("\t//////// Entry Points");
            Dictionary<string, string> entryPoints = GetEntryPoints(element);
            foreach (KeyValuePair<string, string> ep in entryPoints)
            {
                file.WriteLine("\tthis->" + ep.Key + " = new SB::EntryPoint< " + packageName +  "::" + element.Name + " >(");
                file.WriteLine("\t\t\"" + ep.Key + "\",");
                file.WriteLine("\t\t\"" + ep.Value + "\",");
                file.WriteLine("\t\tthis,");
                file.WriteLine("\t\t&" + packageName + "::" + element.Name + "::_" + ep.Key + ");");

                file.WriteLine("\tthis->AddEntryPoint(" + ep.Key + ");");
            }
            file.WriteLine("");

            // References
            ProcessReferences(element);

            // Cointainers
            ProcessContainers(element);

            // Init fields to zero value
            file.WriteLine("\t//////// Initialising fields to zero values");
            List<string> fields = GetFieldsZeroValues(element);
            foreach (string field in fields)
                file.WriteLine("\tthis->" + field + ";");
            fields = GetPortsZeroValues(element);
            foreach (string field in fields)
                file.WriteLine("\tthis->" + field + ";");
            // Init fields to default value
            file.WriteLine("\t//////// Initialising fields to initial values");
            fields = GetFieldsInitValues(element);
            foreach (string field in fields)
                file.WriteLine("\tthis->" + field + ";");
            fields = GetPortsInitValues(element);
            foreach (string field in fields)
                file.WriteLine("\tthis->" + field + ";");
            file.WriteLine("}");
            file.WriteLine("");

            /////////// _Destroy()
            file.WriteLine("void " + packageName + "::" + element.Name + "::_Destroy()\n{");

            foreach (KeyValuePair<string, string> ep in entryPoints)
                file.WriteLine("\tdelete(this->" + ep.Key + ");");
            file.WriteLine("}");
            file.WriteLine("");
        }
        private void ProcessOperations(Element element, string packageName, string type)
        {
            file.WriteLine("/// ------------------------------------------------------");
            if (type == "SMP2entryPoint")
                file.WriteLine("/// --------------------- Entry Points ---------------------");
            else
                file.WriteLine("/// --------------------- Operations ---------------------");
            foreach (Method method in element.Methods)
            {
                if (method.Stereotype != type)
                    continue;
                if(type != "SMP2entryPoint")
                    file.WriteLine(method.ReturnType + " " + packageName + "::" + element.Name + "::" + method.Name + "(" + ProcessParameters(method) + ")" + (method.IsConst ? " const" : ""));
                else
                    file.WriteLine("void " + packageName + "::" + element.Name + "::_" + method.Name + "()");

                file.WriteLine("{");
                file.WriteLine("\t// " + method.Behavior);
                file.WriteLine("}");
                file.WriteLine("");
            }
        }
        
        private string ProcessParameters(Method method)
        {
            string parameters = "";
            foreach (Parameter param in method.Parameters)
            {
                parameters += param.IsConst ? "const " : "";
                parameters += param.Type + " " + param.Name;
            }
            return parameters;
        }
        private Dictionary<string, string> GetEntryPoints(Element element)
        {
            Dictionary<string, string> entryPoints = new Dictionary<string, string>();
            foreach (Method method in element.Methods)
            {
                if (method.Stereotype == "SMP2entryPoint")
                    entryPoints.Add(method.Name, method.Notes);
            }
            return entryPoints;
        }
        private void ProcessReferences(Element element)
        {
            file.WriteLine("\t//////// References");

            // Search for embbeded elements of type Port
            foreach (Element port in element.Elements)
            {
                // Only considerer ports with Required SMP2interfaces
                if (port.Type != "Port" || port.EmbeddedElements.Count <= 0 || port.Stereotype != "SMP2interface")
                    continue;
                foreach (Element expInterface in port.EmbeddedElements)
                {
                    if (expInterface.Type != "RequiredInterface")
                        continue;
                    file.WriteLine("\t_" + port.Name                    
                            + " = SB::Management::ManagedReference< "
                            + expInterface.Name
                            + " >::CreateInstance(");
                    file.WriteLine("\t\t" + "\"" + port.Name + "\" ");
                    file.WriteLine("\t\t, " + "\"" + port.Notes + "\"");
                    file.WriteLine("\t\t, " + "this");
                    if(port.Multiplicity != "")
                    {
                        string[] mult = port.Multiplicity.Replace("..", "#").Split('#');
                        int lower = 0;
                        int upper = -1;
                        Int32.TryParse(mult[0].TrimStart('['), out lower);
                        if (mult.Count() > 1)
                            Int32.TryParse(mult[1].TrimEnd(']').Replace("*", "-1"), out upper);
                        else
                            upper = lower;
                        file.WriteLine("\t\t, " + lower);
                        if(upper > 0)
                            file.WriteLine("\t\t, " + upper);
                    }
                    file.WriteLine("\t);");
                    file.WriteLine("\t" + port.Name + " = _" + port.Name + ";");
                    file.WriteLine("");
                }
            }
            file.WriteLine("");
        }
        private void ProcessContainers(Element element)
        {
            file.WriteLine("\t//////// Containers");
            // Search for composite connectors
            foreach (Connector connector in element.Connectors)
            {
                Element client = Repository.GetElementByID(connector.ClientID);
                if (connector.Type != "Aggregation" || connector.Stereotype != "SMP2container" || element.Name == client.Name)
                    continue;
                
                file.WriteLine("\t_" + connector.ClientEnd.Role 
                        + " = SB::Management::ManagedContainer< "
                        + client.Name
                        + " >::CreateInstance(");
                file.WriteLine("\t\t" + "\"" + connector.ClientEnd.Role + "\" ");
                file.WriteLine("\t\t, " + "\"" + connector.Notes + "\"");
                file.WriteLine("\t\t, " + "this");
                if (connector.ClientEnd.Cardinality != "")
                {
                    string[] mult = connector.ClientEnd.Cardinality.Replace("..", "#").Split('#');
                    int lower = 0;
                    int upper = -1;
                    Int32.TryParse(mult[0].TrimStart('['), out lower);
                    if (mult.Count() > 1)
                        Int32.TryParse(mult[1].TrimEnd(']').Replace("*", "-1"), out upper);
                    else
                        upper = lower;
                    file.WriteLine("\t\t, " + lower);
                    if (upper > 0)
                        file.WriteLine("\t\t, " + upper);
                }
                file.WriteLine("\t);");
                file.WriteLine("\t\t" + connector.ClientEnd.Role + " = _" + connector.ClientEnd.Role + ";");
                file.WriteLine("");
            }
            file.WriteLine("");
        }        
        private List<string> GetFieldsZeroValues(Element element)
        {
            List<string> fields = new List<string>();
            foreach (EA.Attribute att in element.Attributes)
            {
                if (att.Stereotype != "SMP2field")
                    continue;

                string value = UtilClass.GetInitValuebByType(att.Type);
                if (!att.IsCollection)
                    fields.Add(att.Name + " = " + value); // field = value;
                else
                {
                    int upperBound = 0;
                    if (Int32.TryParse(att.UpperBound, out upperBound))
                        for(int i = 0; i < upperBound; i++)
                            fields.Add(att.Name + "[" + i + "] = " + value); // field[i] = value;
                }
            }
            return fields;
        }
        private List<string> GetPortsZeroValues(Element element)
        {
            List<string> fields = new List<string>();
            foreach (Element port in element.EmbeddedElements)
            {
                if (port.Type != "Port")
                    continue;
                if (port.EmbeddedElements.Count > 0) // do not parse ports with embbeded interfaces
                    continue;

                string value = UtilClass.GetInitValuebByType(UtilClass.GetTagValue(port.TaggedValues, "type"));
                string upperBound = ""; // #??# old = UtilClass.GetTagValue(port.TaggedValues, "upperBound");
                upperBound = UtilClass.MultiplicityToUpperBound(port.Multiplicity);
                if (upperBound == "")
                    fields.Add(port.Name + " = " + value); // field = value;
                else
                {
                    int upperBoundValue = 0;
                    if (Int32.TryParse(upperBound, out upperBoundValue))
                        for (int i = 0; i < upperBoundValue; i++)
                            fields.Add(port.Name + "[" + i + "] = " + value); // field[i] = value;
                }
            }
            return fields;
        }       

        private List<string> GetFieldsInitValues(Element element)
        {
            List<string> fields = new List<string>();
            foreach (EA.Attribute att in element.Attributes)
            {
                if (att.Stereotype != "SMP2field" || att.Default == "")
                    continue;
                
                string field = att.Name;

                if (!att.IsCollection)
                    fields.Add(field + " = " + att.Default);
                else
                {                    
                    int upperBoundValue = 0;
                    if (Int32.TryParse(att.UpperBound, out upperBoundValue))
                    {
                        string[] values = att.Default.Split(',');
                        int i = 0;
                        foreach (string value in values)
                        {
                            if (i < upperBoundValue)
                                fields.Add(field + "[" + i + "] = " + value.TrimStart('{').TrimEnd('}'));
                            else
                                MessageBox.Show("Warning: Index of default value for " + att.Name + " exceeds its upper bound!");
                            i++;
                        }
                    }
                }
            }
            return fields;
        }
        private List<string> GetPortsInitValues(Element element)
        {
            List<string> fields = new List<string>();
            foreach (Element port in element.EmbeddedElements)
            {
                if (port.Type != "Port")
                    continue;
                if (port.EmbeddedElements.Count > 0) // do not parse ports with embbeded interfaces
                    continue;

                string value = UtilClass.GetTagValue(port.TaggedValues, "value");
                string upperBound = UtilClass.MultiplicityToUpperBound(port.Multiplicity);  // #??# old UtilClass.GetTagValue(port.TaggedValues, "upperBound");
                /*
                foreach (TaggedValue tag in port.TaggedValues)
                {
                    if (tag.Name == "value")
                        value = tag.Value;
                    else if (tag.Name == "upperBound")
                        upperBound = tag.Value;                    
                } */
                if (value != "")
                {
                    if (upperBound == "")
                        fields.Add(port.Name + " = " + value); // field = value;
                    else
                    {
                        string[] values = value.Split(',');
                        int i = 0;
                        int upperBoundValue = 0;
                        if (Int32.TryParse(upperBound, out upperBoundValue))
                        {
                            foreach (string defaultValue in values)
                            {
                                if (i < upperBoundValue)
                                    fields.Add(port.Name + "[" + i + "] = " + defaultValue.TrimStart('{').TrimEnd('}'));
                                else
                                    MessageBox.Show("Warning: Index of default value for " + port.Name + " exceeds its upper bound!");
                                i++;
                            }
                        }
                    }
                }
            }
            return fields;
        }
        private void ProcessIModelMethods(Element element, string packageName)
        {

            file.WriteLine("/// --------------------------------------------------------");
            file.WriteLine("/// --------------------- IModel ---------------------");

            // Configure
            file.WriteLine("void " + packageName + "::" + element.Name + "::Configure(Smp::Services::ILogger* logger) throw (Smp::IModel::InvalidModelState)");
            file.WriteLine("{");
            file.WriteLine("\t// User code here");
            file.WriteLine("");
            file.WriteLine("\t// At the end call method from base class");
            file.WriteLine("\tSB::Model::Configure(logger);");
            file.WriteLine("}");
            file.WriteLine("");

            // Connect
            file.WriteLine("void " + packageName + "::" + element.Name + "::Connect(Smp::ISimulator* simulator) throw (Smp::IModel::InvalidModelState)");
            file.WriteLine("{");
            file.WriteLine("\t// User code here");
            file.WriteLine("");
            file.WriteLine("\t// At the end call method from base class");
            file.WriteLine("\tSB::Model::Connect(simulator);");
            file.WriteLine("}");
            file.WriteLine("");

            // Publish
            file.WriteLine("void " + packageName + "::" + element.Name + "::Publish(Smp::IPublication* receiver) throw (Smp::IModel::InvalidModelState)");
            file.WriteLine("{");
        
            file.WriteLine("\t// Temp reference to publish operation's parameters");
            file.WriteLine("\tSmp::Publication::IPublishOperation* operation = NULL;");
            file.WriteLine("");
        
            file.WriteLine("\t// Call base implementation first");
            file.WriteLine("\tSB::Model::Publish(receiver);");
            file.WriteLine("");

            if (element.Methods.Count > 0)
            {
                file.WriteLine("\t// ----- Publish Operations");
                foreach (Method op in element.Methods)
                    PublishOperation(op);
                file.WriteLine("");
            }
            
            file.WriteLine("\t// ----- Publish Fields");
            foreach (EA.Attribute att in element.Attributes)
                PublishAttribute(att);
            PublishPortAttributes(element);
            file.WriteLine("");
            file.WriteLine("}");
            file.WriteLine("");
            
        }
        public void ProcessImplementedInterfaces(Element element, string packageName)
        {
            // Search for embbeded elements of type Port
            foreach (Element port in element.Elements)
            {
                // Only considerer ports with Required SMP2interfaces
                if (port.Type != "Port" || port.EmbeddedElements.Count <= 0 || port.Stereotype != "SMP2interface")
                    continue;
                foreach (Element expInterface in port.EmbeddedElements)
                {
                    if (expInterface.Type != "ProvidedInterface")
                        continue;
                    // Search for provided interface in the current package
                    Package currPpackage = Repository.GetPackageByID(element.PackageID);
                    foreach (Element model in currPpackage.Elements)
                    {
                        if (model.Name != expInterface.Name)
                            continue;
                        foreach (Method op in model.Methods)
                        {
                            if (!op.IsPure)
                                continue;

                            string code = op.ReturnType + " " + packageName + "::" + element.Name + "::" + op.Name + "(";
                            foreach (Parameter param in op.Parameters)
                                code += (param.IsConst ? "const " : "") + param.Type + " " + param.Name + ",";
                            code = code.TrimEnd(',');
                            code += ")" + (op.IsConst ? " const" : "");
                            file.WriteLine(code);
                            file.WriteLine("{");
                            file.WriteLine("\t// User code here");
                            if (op.ReturnType != "void")
                                file.WriteLine("\treturn " + op.ReturnType + "()");
                            file.WriteLine("}");
                            file.WriteLine("");
                        }
                    }

                }

            }
        }
        private void PublishOperation(Method op)
        {
            // operation view kind
            string viewKind = UtilClass.GetTagValue(op.TaggedValues, "viewkind").ToLower();
            if (viewKind == "")
                viewKind = "none";

            file.WriteLine("\t// " + op.Name);
            file.WriteLine("\toperation = receiver->PublishOperation(");
            file.WriteLine("\t\t\"" + op.Name + "\",");
            file.WriteLine("\t\t\"" + op.Notes + "\",");
            file.WriteLine("\t\t" + ViewKind[viewKind] + ");");
            file.WriteLine("");

            if (op.Parameters.Count > 0)
            {
                file.WriteLine("\tif(operation) {");
                if (op.ReturnType != "void")
                {
                    file.WriteLine("\t\toperation->PublishParameter(");
                    file.WriteLine("\t\t\t\"Return\",");
                    file.WriteLine("\t\t\t\"Return Parameter\",");
                    file.WriteLine("\t\t\t" + UtilClass.GetTypeUuid(op.ReturnType) + ",");
                    file.WriteLine("\t\t\t Smp::Publication::PDK_Return);");
                    file.WriteLine("");
                }
                foreach (Parameter param in op.Parameters)
                {
                    file.WriteLine("\t\toperation->PublishParameter(");
                    file.WriteLine("\t\t\t\"" + param.Name + "\",");
                    file.WriteLine("\t\t\t\"" + param.Notes + "\",");
                    file.WriteLine("\t\t\t" + UtilClass.GetTypeUuid(param.Type) + ",");
                    file.WriteLine("\t\t\t" + ParameterDirectionKind[param.Kind] + ");");
                    file.WriteLine("");
                }
                
                file.WriteLine("\t}");
                file.WriteLine("");
            }
        }
        private void PublishAttribute(EA.Attribute att)
        {
            // attribute view kind
            string viewKind = UtilClass.GetTagValue(att.TaggedValues, "viewKind").ToLower();
            if (viewKind == "")
                viewKind = "none";
            string state = UtilClass.GetTagValue(att.TaggedValues, "state").ToLower();
            if(state == "")
                state = "true";

            file.WriteLine("\t// " + att.Name);
            
            if (att.IsCollection)
            {
                int upperBoundValue = 0;
                Int32.TryParse(att.UpperBound, out upperBoundValue);
                file.WriteLine("\treceiver->PublishArray(");
                file.WriteLine("\t\t\"" + att.Name + "\",");
                file.WriteLine("\t\t\"" + att.Notes + "\",");
                file.WriteLine("\t\t" + upperBoundValue + ",");
                file.WriteLine("\t\tstatic_cast<void*>(" + att.Name + "),");
                file.WriteLine("\t\t" + UtilClass.GetPublicationType(att.Type) + ",");
            }
            else
            {
                file.WriteLine("\treceiver->PublishField(");
                file.WriteLine("\t\t\"" + att.Name + "\",");
                file.WriteLine("\t\t\"" + att.Notes + "\",");
                if (att.Type == "Smp::String8" || att.Type == "Smp::Duration" || att.Type == "Smp::DateTime")
                {
                    file.WriteLine("\t\tstatic_const<void*>(&" + att.Name + "),");
                    file.WriteLine("\t\t" + UtilClass.GetPublicationType(att.Type) + ",");
                }
                else
                    file.WriteLine("\t\t&" + att.Name + ",");
            }
            file.WriteLine("\t\t" + ViewKind[viewKind] + "," );
            file.WriteLine("\t\t" + state + ",");
            file.WriteLine("\t\tfalse,");
            file.WriteLine("\t\tfalse);");
            file.WriteLine("");
        }
        private void PublishPortAttributes(Element element)
        {
            foreach (Element port in element.EmbeddedElements)
            {
                if (port.Type != "Port")
                    continue;
                if (port.EmbeddedElements.Count > 0) // do not parse ports with embbeded interfaces
                    continue;

                string output = UtilClass.GetTagValue(port.TaggedValues, "output").ToLower();
                output = (output == "" ? "false" : output);
                string input = UtilClass.GetTagValue(port.TaggedValues, "input").ToLower();
                input = (input == "" ? "false" : input);

                if (input == "false" && output == "false")
                    continue;

                string value = UtilClass.GetTagValue(port.TaggedValues, "value");
                string type = UtilClass.GetTagValue(port.TaggedValues, "type");
                string viewKind = UtilClass.GetTagValue(port.TaggedValues, "viewKind").ToLower();
                viewKind = (viewKind == "" ? "none" : viewKind);
                string state = UtilClass.GetTagValue(port.TaggedValues, "state").ToLower();
                state = (state == "" ? "true" : state);
                string upperBound = UtilClass.MultiplicityToUpperBound(port.Multiplicity);  // #??# old UtilClass.GetTagValue(port.TaggedValues, "upperBound");

                file.WriteLine("\t// " + port.Name);

                if (upperBound != "")
                {
                    int upperBoundValue = 0;
                    Int32.TryParse(upperBound, out upperBoundValue);
                    file.WriteLine("\treceiver->PublishArray(");
                    file.WriteLine("\t\t\"" + port.Name + "\",");
                    file.WriteLine("\t\t\"" + port.Notes + "\",");
                    file.WriteLine("\t\t" + upperBoundValue + ",");
                    file.WriteLine("\t\tstatic_cast<void*>(" + port.Name + "),");
                    file.WriteLine("\t\t" + UtilClass.GetPublicationType(type) + ",");                    
                }
                else
                {
                    file.WriteLine("\treceiver->PublishField(");
                    file.WriteLine("\t\t\"" + port.Name + "\",");
                    file.WriteLine("\t\t\"" + port.Notes + "\",");
                    file.WriteLine("\t\t&" + port.Name + ",");
                }
                file.WriteLine("\t\t" + ViewKind[viewKind] + ",");
                file.WriteLine("\t\t" + state + ",");
                file.WriteLine("\t\t" + input + ",");
                file.WriteLine("\t\t" + output + ");");
                file.WriteLine("");
            }
        }
        
    } // End class
} // end namespace
